#include <u.h>
#include <libc.h>
#include <mp.h>
#include <libsec.h>
#include <bio.h>

enum {
	MaxBlocks = 0x7fffffff,
	OverflowBufSize = 0xFF,
};

struct blocksum {
	ulong	blocknum;
	u32int	rolling;
	uchar	md4[MD4dlen];
};

static int verbose = 0;
static unsigned blocksize = 2048;

static ulong stat_blockmatch = 0;
static ulong stat_rollingmiss = 0;

/* 
 * rsync rolling checksum.
*/
u32int
get_rolling (char *buf1, vlong len)
{
	vlong 	i;
	u32int	s1, s2;
	
	signed char *buf = (signed char *)buf1;

	s1 = s2 = 0;
	for (i = 0; i < (len-4); i += 4) {
		s2 += 4*(s1 + buf[i]) + 3*buf[i+1] + 2*buf[i+2] + buf[i+3];
		s1 += buf[i+0] + buf[i+1] + buf[i+2] + buf[i+3];
	}
	for (; i < len; i++) {
		s1 += buf[i]; s2 += s1;
	}
	return (s1 & 0xffff) + (s2 << 16);
}

int
sumcompare(void *a1, void*b1)
{
	struct blocksum *a;
	struct blocksum *b;

	a = (struct blocksum *) a1;
	b = (struct blocksum *) b1;

	if (a->rolling < b->rolling) {
		return -1;
	} else if (a->rolling > b->rolling) {
		return 1;
	} else {
		return 0;
	}
}

int
buildsums(int fdin, struct blocksum *sums, ulong sumcount, uchar *wholemd4)
{
	char *buf;
	struct blocksum *s;
	DigestState *wholefile = nil;
	uint		i;
	uint		count;

	buf = malloc(blocksize);
	if (!buf) {
		fprint (2, "error: out of memory allocating block buffer: %r");
		exits("couldn't allocate read buffer");
	}

	for (i = 0, s=sums; i < sumcount && ((count = readn(fdin, buf, blocksize)) > 0); i++, s++) {
		s->blocknum = i;
		s->rolling = get_rolling(buf, count);
		md4((uchar *)buf, count, s->md4, nil);
		wholefile = md4((uchar *)buf, count, nil, wholefile);
	}
	if (wholemd4) {
		md4(nil, 0, wholemd4, wholefile);
	}

	free(buf);

	return i;
}

struct blocksum *
findblocksum(u32int rolling, struct blocksum *s, ulong len)
{
	int		n;

	n = len  / 2;

	if (s[n].rolling == rolling) {
		return s+n;
	} else if (s[n].rolling > rolling) {
		if (n > 0) {
			return findblocksum(rolling, s, n-1);
		}
	} else if (s[n].rolling < rolling) {
		if (n < len) {
			return findblocksum(rolling, s+n+1, len-n-1);
		}
	}
	return nil;
}

void
emitblock(ulong blocknum)
{
	fprint(1, "useblock %08ulX\n", blocknum);
}

void
emitverbatim(char *buf, int len)
{
	int i;

	for (i = 0; i < len-256; i += 256) {
		fprint(1, "raw %02X %.*[\n", 256, 256, buf+i);
	}
	fprint(1, "raw %02X %.*[\n", len-i, len-i, buf+i);
}

void
blockscan(int fd, struct blocksum *sums, ulong sumcount, uchar *wholemd4)
{
	char *buf;
	ulong bufofs;
	char overflow[OverflowBufSize];
	int overflowidx = 0;
	struct blocksum *	b;
	DigestState *s = nil;
	int count;
	u32int s1;
	u32int s2;
	u32int sum;
	uchar	bufmd4[MD4dlen];
	Biobuf		bfd;

	/* first up, initialise the buffer */
	buf = malloc(blocksize * 2);
	if (!buf) {
		fprint (2, "error: out of memory allocating block buffer: %r");
		exits("couldn't allocate read buffer");
	}

	/* attach fd to a bio */
	Binit(&bfd, fd, OREAD);

	while (1) {
		/* prefill, abort early if short */
		bufofs = 0;
		count = Bread(&bfd, buf, blocksize);
		if (count < blocksize) {
			goto flushbuffers;
		}

		/* build our initial checkum */
		sum = get_rolling(buf+bufofs, count);

		/* start the scanning process */
		goto blockscan;
		while (!b) {
			/* update checksum */
			s1 = sum & 0xFFFF;
			s2 = sum >> 16;
			s1 -= buf[bufofs];
			s2 -= blocksize*buf[bufofs];

			/* append the byte to the overflow buffer */
			overflow[overflowidx++] = buf[bufofs];
			/* check for overflows in the overflow buffer */
			if (overflowidx >= OverflowBufSize) {
				emitverbatim (overflow, overflowidx);
				s = md4((uchar *)overflow, overflowidx, nil, s);
				overflowidx = 0;
			}

			/* shuffle buffer */
			bufofs++;
			if (bufofs >= blocksize) {
				memmove(buf, buf+bufofs, blocksize-1);
				bufofs = 0;
			}

			/* get next byte (if we can) */
			if (0 >= Bread(&bfd, buf+bufofs+blocksize-1, 1)) {
				/* except we're out of buffer.  d'oh. */
				goto flushbuffers;
			}

			/* update checksum */
			s1 += buf[bufofs+blocksize-1];
			s2 += s1;
			sum = s1 & 0xFFFF | s2 << 16;

			/* search again */
blockscan:
			b = findblocksum(sum, sums, sumcount);
			if (b) {
				md4((uchar *)(buf+bufofs), count, bufmd4, nil);
				if (memcmp(b->md4, bufmd4, MD4dlen)) {
					b = nil;
					stat_rollingmiss++;
				}
			}		
		}
		/* we've matched a block, emit the overflow buffer first */
		if (overflowidx) {
			emitverbatim (overflow, overflowidx);
			s = md4((uchar *)overflow, overflowidx, nil, s);
			overflowidx = 0;
		}
		/* then the block */
		stat_blockmatch++;
		emitblock (b->blocknum);
		s = md4((uchar *)buf+bufofs, count, nil, s);
		count = 0;
	}
flushbuffers:
	if (overflowidx) {
		emitverbatim (overflow, overflowidx);
		s = md4((uchar *)overflow, overflowidx, nil, s);
		overflowidx = 0;
	}
	if (count) {
		emitverbatim(buf+bufofs, count);
		s = md4((uchar *)buf+bufofs, count, nil, s);
		count = 0;
	}
	if (wholemd4)
		md4(nil, 0, wholemd4, s);
	Bterm(&bfd);
	return;	
}

void
main(int argc, char *argv[])
{
	int	ffd;
	char *f;
	ulong i;
	Dir *finfo;
	uchar	srcsum[2][MD4dlen];
	struct blocksum *	srcblocksums;
	ulong	blocksumcount;

	/* install fmt for hex buffers and base64 */
	fmtinstall('H', encodefmt);
	fmtinstall('[', encodefmt);

	ARGBEGIN {
	case 'v':
		verbose = 1;
		break;
	case 'b':
		if (!(f = ARGF()))
			goto usage;
		blocksize = atoi(f);
		break;
	default:
		goto usage;
	} ARGEND

	if (argc != 2)
		goto usage;

	ffd = open(argv[0], OREAD);
	if (ffd < 0) {
		fprint(2, "cp: can't open %s: %r\n", argv[0]);
		exits("can't open srcfile1");
	}

	/* verify the block size */
	finfo = dirfstat(ffd);
	if ((finfo->length / (vlong)blocksize) > MaxBlocks) {
		fprint(2, "blocksize %d is too small for the source file.\n", blocksize);
		exits("specified blocksize too small");
	}

	/* calculate the number of blocks required */
	blocksumcount = finfo->length / blocksize;
	srcblocksums = calloc(sizeof(struct blocksum), blocksumcount);
	if (!srcblocksums) {
		fprint(2, "out of memory allocating blocksum memory.\n");
		exits("out of memory");
	}
	/* release the dir structure */
	free (finfo); finfo = nil;

	/* and calculate the blocks */
	blocksumcount = buildsums(ffd, srcblocksums, blocksumcount, srcsum[0]);

	/* release the first file */
	close (ffd);

	/* dump the first bit of the bindiff */
	fprint(1, "bindiff 0\n");	/* version number */
	fprint(1, "blocksize %uX\n", blocksize);
	fprint(1, "srcfile %.*[\n", MD4dlen, srcsum[0]);
	fprint(1, "blockcount %ulX\n", blocksumcount);
	for (i = 0; i < blocksumcount; i++) {
		fprint(1, "block %08ulX %08ulX %.*[\n",
			srcblocksums[i].blocknum, (ulong) srcblocksums[i].rolling, 
			MD4dlen, srcblocksums[i].md4);
	}

	/* qsort the checksums so we can use binary search to locate the bits */
	qsort(srcblocksums, blocksumcount, sizeof(*srcblocksums), sumcompare);

	/* open the second file */
	ffd = open(argv[1], OREAD);
	if (ffd < 0) {
		fprint(2, "cp: can't open %s: %r\n", argv[1]);
		exits("can't open srcfile2");
	}

	blockscan(ffd, srcblocksums, blocksumcount, srcsum[1]);

	close(ffd);

	fprint(1, "destfile %.*[\n", MD4dlen, srcsum[1]);

	if (verbose) {
		fprint(2, "Completed:  %uld block hits.  %uld rolling csum collisions\n", 
			stat_blockmatch, stat_rollingmiss);
	}

	exits(nil);
usage:
	fprint(2, "usage:\tbindiff [-v] [-b blocksize] srcfile1 srcfile2\n");
	exits("usage");
}